from marshmallow import fields
import random

model = {
    "fruit": {
        "field": fields.String(required=True),
        "stream": {
            "type": "choice",
            "default": ['tangerine', 'tango', 'turtles'],
            "include": True,
            "existing": True,
        },
    },
    "item": {
        "field": fields.Integer(required=True),
        "stream": {
            "type": "increment",
            "default": 0,
            "include": True,
            "existing": True,
        },
    },
    "price": {
        # "field": fields.Float(missing=random.randint(2, 200)),
        "stream": {
            "type": "increment",
            "default": 69.69,
            "include": True,
            "existing": False,
        },
    },
}


import logging
import time

logging.basicConfig(level=logging.DEBUG)
"""I think this will become an object that holds its state of things
like frequency, and event scheme, which can be accessed and set

This was overly complex with multi threads per worker, but each engine 
will be simple to start, espcielly in this explore stage, so simplfiy and 
take out second layer of threading.

"""


class Engine:

    """blueprint for an engine, which is passed a domain class instance to call for data"""

    def __init__(self, domain, sio_app) -> None:
        self.domain = domain
        self.sio = sio_app

        self.frequency = 1.0
        self.run = True

        self.limit_mode = False
        self.limit = 10
        self.limit_counter = 0

        self.burst_mode = False
        self.burst_limit = 30
        self.burst_counter = 0
        self.burst_frequency = 0.2

        logging.info(f"engine domain object {self.domain.name} is {self.domain}")

        # self.generate()  # to auto start, is this legit to call here?

    def collect_emit(self):
        """collects new event data from the passed domain instance and emits event"""

        event = self.domain.new_event()
        logging.info(f"engine called domain {event}")
        self.sio.emit("stream", data=event)

    def generate(self):
        """generates new data and emits"""

        while self.run == True:
            if self.limit_mode:
                if self.limit_counter < self.limit:

                    self.collect_emit()
                    self.limit_counter += 1
                else:
                    self.stop()

                time.sleep(self.frequency)

            if self.burst_mode:
                if self.burst_counter < self.burst_limit:
                    self.collect_emit()
                    self.burst_counter += 1
                else:
                    self.burst_mode = False
                    self.burst_counter = 0
                time.sleep(self.burst_frequency)

            else:
                self.collect_emit()
                time.sleep(self.frequency)

    def set_frequency(self, new_freq):
        """Setter for frequency"""
        self.frequency = new_freq

    def set_burst(self):
        """setter to start a burst"""
        self.burst_mode = True

    def set_error_mode_on(self):
        """setter to set error mode for domain to on"""
        self.domain.error_mode = True

    def set_error_mode_off(self):
        """setter to set error mode for domain to off"""
        self.domain.error_mode = False

    def stop(self):
        """set self.run to False and stop engine"""
        self.run = False

    def start(self):
        """set self.run to True and start engine"""
        self.run = True
        self.limit_counter = 0
        self.sio.start_background_task(self.generate)

    def burst(self):
        """trigger burst mode"""



from flask import Flask, jsonify, request, send_file
from flask_socketio import SocketIO

import random
import logging

logging.basicConfig(level=logging.INFO)


from ..engine import Engine
from ..domains import Domain


app = Flask("hw-server")
sio = SocketIO(app)


@app.get("/")
def index():
    return jsonify(server=f"says hello and {random.random()}")


@app.get("/start")
def start():
    engine = random.choice(engines)
    engine.start()

    return jsonify(server=f"started engine {engine.domain.name}")


@app.get("/stop")
def stop():
    engine = random.choice(engines)
    engine.stop()

    return jsonify(server=f"stopped engine {engine.domain.name}")


@app.get("/frequency")
def command():
    """this could be one route to test in operation param changes across the command_q"""
    new_freq = random.randint(1, 6)

    # so the domain of the engine can be idenfitief here with
    # for engine in engines: if engine.domain == xyz then do soemthing
    engine = random.choice(engines)
    engine.set_frequency(new_freq)
    return jsonify(server=f"adjusted engine {engine.domain.name} freq to {new_freq}")


@app.get("/burst")
def burst():
    engine = random.choice(engines)
    engine.set_burst()

    return jsonify(
        server=f"initiated burst for engine {engine.domain.name} with {engine.burst_limit}"
    )


@app.get("/error_on")
def error_on():
    engine = random.choice(engines)
    engine.set_error_mode_on()

    return jsonify(server=f"error mode set for engine {engine.domain.name}")


@app.post("/add_field")
def add_word():

    data = request.json

    this_domain = data['domain']

    r = "huh"
    for domain in domains:
        if this_domain == domain.name:
            r = domain.set_field(data)
            break

    return jsonify(server=r)

@app.route('/ui', defaults={'path': ''})
@app.route('/<path:path>')
def catch_all(path):
    print(path)
    if path.endswith('.js'):
        return send_file(f"src/headwaters/{path}", mimetype="text/javascript")
    elif path.endswith('.css'):
        return send_file(f"src/headwaters/{path}", mimetype="text/css")
    elif path.endswith('.ico'):
        return send_file(f"src/headwaters/{path}")
    elif path.endswith('.svg'):
        return send_file(f"src/headwaters/{path}")
    else:
        return send_file("src/headwaters/ui/index.html", mimetype="text/html")

@sio.event("connect")
def connect_hndlr():
    logging.info(f"sio conneciton rcvd {sio.sid}")


engines = []
domains = []


def run(selected_domains):
    """

    what about: the Engine instances are created, then the generate method of each is multithreaded...
    could the thread have access to the object properties in that thread? NO surely not
    THIS SEEMS TO WORK?!?!?

    ALSO CAN I KNOW INSTNATIATE ENGINE OBKECT OUTWITH SERVER
    CAN I TEST THEM WITHOUT USING HTTP REQUESTS??

    now the thinking is to have the domain classed available in the server scope
    that way i can access the new_event() method of a domain instance fromt he generator thread and
    also access it for CUD ops

    """

    for selected_domain in selected_domains:
        domain = Domain(selected_domain)
        domains.append(domain)
        engines.append(Engine(domain, sio))

    for engine in engines:
        sio.start_background_task(target=engine.generate)

    sio.run(app, debug=False)



""" cli.py is the main entry point for the app and the top level """

import click
import logging

from . import server

# TODO need to be very clear in docs how to pass a list to the cli
# i have forgotten right now and can't start multiple
# hw -e fast -e slow etc...
@click.command()
@click.option(
    "--domains",
    "-d",
    default=["fruits"],
    multiple=True,
    help="specify the domain(s) for the server, with each domain preceded by a -d",
)
def main(domains: str) -> None:

    server.run(domains)

if __name__ == "__main__":
    main()



import pandas as pd
import random
from marshmallow import Schema, ValidationError
import logging

from .fruits import data as fruits_data
from .fruits import model as fruits_model

class Domain:

    """
    This needs to be passed just a string indictaing the domain type,
    and the instance needs to grab the data and model from the right place
    using the pckg data

    this needs the getters and setter for data to be added
    """
    def __init__(self, domain):
        self.name = domain
        if self.name == 'fruits':
            self.model = fruits_model
        logging.info(self.model)
        
        self.data = fruits_data
        logging.info(self.data)

        self.new_data = [] # holding solution for the expanign choice issue

        self.process_passed_data()

        if not self.validate_data():
            raise ValueError(f"model to data validation error")

    def process_passed_data(self):
        """use pandas to comvert any passed data shape to a nice shape"""
        key_list = []

        for k in self.model.keys():
            if self.model[k]["stream"]["existing"]:
                key_list.append(k)

        columns = key_list

        df = pd.DataFrame(data=self.data, columns=columns)

        self.data = df.to_dict(orient="records")

    def create_data_schema(self):
        """create a marshmallow schema from passed model looking for data def only
        only therefore applies to data that exists in the data files, so can exlude
        existing: False
        
        """
        d = {}
        for k in self.model.keys():
            if self.model[k]['stream']['existing']:
                d.update({k: self.model[k]["field"]})

        DataSchema = Schema.from_dict(d)
        return DataSchema

    def validate_data(self):
        """
        use create marshmallow instance to validate passed data
        """

        DataSchema = self.create_data_schema()

        try:

            DataSchema(many=True).load(self.data)
            return True
        except ValidationError as e:
            logging.info(e)
            return False

    def new_event(self):
        """Once loaded, shaped and validated against the model this method can then be safely called
        
        the stream part of the model is defo going to need a schema validator: ie
        the actual generator types and setting needs a lot of work also
        if type == choice then default mut be list of len min 1 etc etc
        if exsiting == True, then must have field definition
        """

        new_event = {}

        for k in self.model.keys():

            field = self.model[k]["stream"]
            default = self.model[k]["stream"]["default"]
            if field["include"]:
                if field["type"] == "choice":
                    try:
                        new_event[k] = random.choice(self.data)[k]
                    except KeyError:
                        """ handles where an event field not in the original data is being picked from"""
                        new_event[k] = random.choice(default)
                if field["type"] == "increment":
                    try:
                        new_event[k] = self.new_data[-1][k] + 1
                    except:
                        new_event[k] = default  
                if field["type"] == "infer":
                    new_event[k] = "inference function called here"
                
        self.new_data.append(new_event)
        return new_event

    def set_field(self, data):
        """ setter to add a new field to running domain instance """
        new_field = data['new_field']
        self.model[new_field] = data['settings']

        return "set_field complete"